from .constants_of_motion import *
from numpy import cos, sin

class KerrSpacetime:
    """
        Class representing a black hole with mass M and spin parameter a

        :param a: dimensionless angular momentum
        :type a: double
        :param M: mass of the black hole
        :type M: double

        :ivar a: dimensionless angular momentum
        :ivar M: mass of the black hole
    """
    def __init__(self,a,M):
        self.a = a
        self.M = M

    def separatrix(self,e,x):
        """
        Computes the value of p at the separatrix for a given value of e and x

        :param e: orbital eccentricity
        :type e: double
        :param x: cosine of the orbital inclination
        :type x: double

        :rtype: double
        """
        return separatrix(self.a,e,x)
    
    def is_stable(self,orbit):
        """
        Determines whether a given orbit is stable or not.

        :param orbit: orbit to be tested
        :type orbit: Orbit

        :rtype: bool
        """
        return is_stable(self.a,orbit.p,orbit.e,orbit.x)
    
    def inner_horizon(self):
        """
        Computes the radius of the inner event horizon

        :return: dimensionless radius of the inner event horizon
        :rtype: double
        """
        return 1-sqrt(1-self.a**2)
    
    def outer_horizon(self):
        """
        Computes the radius of the outer event horizon

        :return: dimensionless radius of the outer event horizon
        :rtype: double
        """
        return 1+sqrt(1-self.a**2)
    
    def metric(self,t,r,theta,phi):
        """
        Returns the matrix representation of the metric at a given point expressed in Boyer-Lindquist coordinates.

        :param t: time coordinate
        :type t: double
        :param r: radial coordinate
        :type r: double
        :param theta: polar coordinate
        :type theta: double
        :param phi: azimuthal coordinate
        :type phi: double

        :rtype: numpy.ndarray
        """
        a = self.a
        sigma = r**2+a**2*cos(theta)**2
        delta = r**2-2*r+a**2
        return np.array(
            [[-(1-2*r/sigma),               0,              0,      -2*a*r*sin(theta)**2/sigma],
             [0,                            sigma/delta,    0,      0],
             [0,                            0,              sigma,  0],
             [-2*a*r*sin(theta)**2/sigma,   0,              0,      sin(theta)**2*(r**2+a**2+2*a**2*r*sin(theta)**2/sigma)]
             ]
        )
    
    def norm(self,t,r,theta,phi,v):
        """
        Computes the norm of a vector at a given point in spacetime expressed in Boyer-Lindquist coordinates

        :param t: time coordinate
        :type t: double
        :param r: radial coordinate
        :type r: double
        :param theta: polar coordinate
        :type theta: double
        :param phi: azimuthal coordinate
        :type phi: double
        :param v: vector to compute the norm of
        :type v: array_like

        :return: norm of v
        :rtype: double
        """
        return np.dot(v,np.dot(self.metric(t,r,theta,phi),v))
    
    def four_velocity(self,t,r,theta,phi,constants):
        r"""
        Computes the four velocity of a given trajectory

        :param t: time component of trajectory
        :type t: function
        :param r: radial component of trajectory
        :type r: function
        :param theta: polar component of trajectory
        :type theta: function
        :param phi: azimuthal component of trajectory
        :type phi: function
        :param constants: tuple of constants of motion in the form :math:`(E,L,Q)`
        :type constants: tuple(double, double, double)

        :return: components of the four velocity (i.e. :math:`(\frac{dt}{d\tau}, \frac{dr}{d\tau},\frac{d\theta}{d\tau},\frac{d\phi}{d\tau})`)
        :rtype: tuple(function, function, function, function)
        """
        a = self.a
        E, L, Q = constants
        # radial polynomial
        R = lambda r: (E*(r**2+a**2)-a*L)**2-(r**2-2*r+a**2)*(r**2+(a*E-L)**2+Q)
        # polar polynomial
        Z = lambda z: Q-(Q+a**2*(1-E**2)+L**2)*z**2+a**2*(1-E**2)*z**4

        def t_prime(time):
            delta = r(time)**2-2*r(time)+a**2
            sigma = r(time)**2+a**2*cos(theta(time))**2
            return 1/sigma*((r(time)**2+a**2)/delta*(E*(r(time)**2+a**2)-a*L)-a**2*E*(1-cos(theta(time))**2)+a*L)

        def r_prime(time):
            sigma = r(time)**2+a**2*cos(theta(time))**2
            return sqrt(R(r(time)))/sigma

        def theta_prime(time):
            sigma = r(time)**2+a**2*cos(theta(time))**2
            return sqrt(Z(cos(theta(time))))/sigma
        
        def phi_prime(time):
            sigma = r(time)**2+a**2*cos(theta(time))**2
            delta = r(time)**2-2*r(time)+a**2
            return 1/sigma*(a/delta*(E*(r(time)**2+a**2)-a*L)+L/(1-cos(theta(time))**2)-a*E)
        
        return t_prime, r_prime, theta_prime, phi_prime



from .plunging_solutions import *
from .plunging_solutions import _plunging_radial_roots
from .bound_solutions import *
from .orbit import Orbit
import matplotlib.pyplot as plt

class PlungingOrbit(Orbit):
    """
    Class representing a plunging orbit

    :param a: dimensionaless spin parameter
    :type a: double
    :param E: dimensionless energy
    :type E: double
    :param L: dimensionless angular momentum
    :type L: double
    :param Q: dimensionless Carter constant
    :type Q: double

    :ivar a: dimensionaless spin parameter
    :ivar E: dimensionless energy
    :ivar L: dimensionless angular momentum
    :ivar Q: dimensionless Carter constant
    :ivar upsilon_r: radial Mino frequency
    :ivar upsilon_theta: polar Mino frequency
    """
    def __init__(self,a,E,L,Q):
        self.a, self.E, self.L, self.Q = a, E, L, Q
        self.upsilon_r, self.upsilon_theta = plunging_mino_frequencies(a,E,L,Q)

    def trajectory(self,initial_phases=(0,0,0,0)):
        r"""
        Computes the components of the trajectory as a function of mino time

        :param initial_phases: tuple of initial phases, defaults to (0,0,0,0)
        :type initial_phases: tuple, optional
        
        :return: tuple of functions :math:`t(\lambda), r(\lambda), \theta(\lambda), \phi(\lambda)`
        :rtype: tuple(function,function,function,function)
        """
        a, E, L, Q = self.a, self.E, self.L, self.Q
        radial_roots = _plunging_radial_roots(a,E,L,Q)

        if np.iscomplex(radial_roots[3]):
            return self._complex_trajectory(initial_phases)
        else:
            return self._real_trajectory(initial_phases)
    
    def _complex_trajectory(self,initial_phases=(0,0,0,0)):
        """
        Computes the components of the trajectory in the case of two complex and two real radial roots

        :param initial_phases: tuple of initial phases, defaults to (0,0,0,0)
        :type initial_phases: tuple, optional

        :rtype: tuple(function,function,function,function)
        """
        a, E, L, Q = self.a, self.E, self.L, self.Q
        upsilon_r, upsilon_theta = self.upsilon_r, self.upsilon_theta
        r_phases, t_r, phi_r = plunging_radial_solutions_complex(a,E,L,Q)
        theta_phases, t_theta, phi_theta = plunging_polar_solutions(a,E,L,Q)
        q_t0, q_r0, q_theta0, q_phi0 = initial_phases

        # Calculate normalization constants so that t = 0 and phi = 0 at lambda = 0 when q_t0 = 0 and q_phi0 = 0 
        C_t = t_r(q_r0)+t_theta(q_theta0)
        C_phi= phi_r(q_r0)+phi_theta(q_theta0)

        def t(mino_time):
            return t_r(upsilon_r*mino_time+q_r0) + t_theta(upsilon_theta*mino_time+q_theta0) + a*L*mino_time - C_t + q_t0
        
        def r(mino_time):
            return r_phases(upsilon_r*mino_time+q_r0)
        
        def theta(mino_time):
            return theta_phases(upsilon_theta*mino_time+q_theta0)
        
        def phi(mino_time):
            return phi_r(upsilon_r*mino_time+q_r0) + phi_theta(upsilon_theta*mino_time+q_theta0) - a*E*mino_time - C_phi + q_phi0
        
        return t, r, theta, phi
    
    def _real_trajectory(self,initial_phases=(0,0,0,0)):
        """
        Computes the components of the trajectory in the case of four real radial roots

        :param initial_phases: tuple of initial phases, defaults to (0,0,0,0)
        :type initial_phases: tuple, optional

        :rtype: tuple(function,function,function,function)
        """
        a, E, L, Q = self.a, self.E, self.L, self.Q
        constants = (E,L,Q)
        Z = Polynomial([Q,-(Q+a**2*(1-E**2)+L**2),a**2*(1-E**2)])
        radial_roots = _plunging_radial_roots(a,E,L,Q)
        polar_roots = Z.roots()

        upsilon_r, upsilon_theta, upsilon_phi, gamma = mino_frequencies_from_constants(a,constants,radial_roots,polar_roots)
        r_phases, t_r, phi_r = radial_solutions(a,constants,radial_roots)
        theta_phases, t_theta, phi_theta = polar_solutions(a,constants,polar_roots)
        q_t0, q_r0, q_theta0, q_phi0 = initial_phases

        # Calculate normalization constants so that t = 0 and phi = 0 at lambda = 0 when q_t0 = 0 and q_phi0 = 0 
        C_t = t_r(q_r0)+t_theta(q_theta0)
        C_phi= phi_r(q_r0)+phi_theta(q_theta0)

        def t(mino_time):
            # equation 6
            return q_t0 + gamma*mino_time + t_r(upsilon_r*mino_time+q_r0) + t_theta(upsilon_theta*mino_time+q_theta0) - C_t
        
        def r(mino_time):
            return r_phases(upsilon_r*mino_time+q_r0)
        
        def theta(mino_time):
            return theta_phases(upsilon_theta*mino_time+q_theta0)
        
        def phi(mino_time):
            # equation 6
            return q_phi0 + upsilon_phi*mino_time + phi_r(upsilon_r*mino_time+q_r0) + phi_theta(upsilon_theta*mino_time+q_theta0) - C_phi
        
        return t, r, theta, phi


# https://en.wikipedia.org/wiki/Geometrized_unit_system

# speed of light in m/s
c = 299792458.0
# gravitational constant in m^3 kg^-1 s^-2
G = 6.67408e-11
# solar mass in kg
solar_mass = 1.98847e30

def mass_in_kg(M):
    """
    Converts solar masses to kg

    :param M: mass in solar masses
    :type M: double

    :return: mass in kg
    :rtype: double
    """
    return M*solar_mass

def distance_in_meters(d,M):
    """
    Converts distance in geometrized units to meters

    :param d: distance in multiples of M
    :type d: double
    :param M: mass of the primary body in solar masses
    :type M: double

    :return: distance in meters
    :rtype: double
    """
    return d*G/c**2*mass_in_kg(M)

def distance_in_cm(d,M):
    """
    Converts distance in geometrized units to centimeters

    :param d: distance in multiples of M
    :type d: double
    :param M: mass of the primary body in solar masses
    :type M: double

    :return: distance in centimeters
    :rtype: double
    """
    return distance_in_meters(d,M)*100

def distance_in_km(d,M):
    """
    Converts distance in geometrized units to kilometers

    :param d: distance in multiples of M
    :type d: double

    :return: distance in kilometers
    :rtype: double
    """
    return distance_in_meters(d,M)/1000

def distance_in_lightyears(d,M):
    """
    Converts distance in geometrized units to lightyears

    :param d: distance in multiples of M
    :type d: double
    :param M: mass of the primary body in solar masses
    :type M: double

    :return: distance in lightyears
    :rtype: double
    """
    return distance_in_meters(d,M)/9.461e15

def distance_in_au(d,M):
    """
    Converts distance in geometrized units to astronomical units

    :param d: distance in multiples of M
    :type d: double
    :param M: mass of the primary body in solar masses
    :type M: double

    :return: distance in astronomical units
    :rtype: double
    """
    return distance_in_meters(d,M)/1.496e11

def time_in_seconds(t,M):
    """
    Converts time in geometrized units to seconds

    :param t: time in multiples of M
    :type t: double
    :param M: mass of the primary body in solar masses
    :type M: double

    :return: time in seconds
    :rtype: double
    """
    return t*G/c**3*mass_in_kg(M)

def time2_in_seconds2(t,M):
    """
    Converts time^2 in geometrized units to seconds^2

    :param t: time^2 in multiples of M^2
    :type t: double
    :param M: mass of the primary body in solar masses
    :type M: double
    """
    return t*(G/c**3)**2*mass_in_kg(M)**2

def time_in_days(t,M):
    """
    Converts time in geometrized units to days

    :param t: time in multiples of M
    :type t: double
    :param M: mass of the primary body in solar masses
    :type M: double
    :return: time in days
    :rtype: double
    """
    return time_in_seconds(t,M)/86400

def energy_in_joules(E,M):
    """
    Converts energy in geometrized units to joules

    :param E: energy in multiples of M
    :type E: double
    :param M: mass of the primary body in solar masses
    :type M: double

    :return: energy in joules
    :rtype: double
    """
    return E*c**2*mass_in_kg(M)

def energy_in_ergs(E,M):
    """
    Converts energy in geometrized units to ergs

    :param E: energy in multiples of M
    :type E: double
    :param M: mass of the primary body in solar masses
    :type M: double

    :return: energy in ergs
    :rtype: double
    """
    return energy_in_joules(E,M)*1e7

def angular_momentum_in_mks(L,M):
    """
    Converts angular momentum in geometrized units to kg m^2 s^-1

    :param L: angular momentum in multiples of M^2
    :type L: double
    :param M: mass of the primary body in solar masses
    :type M: double

    :return: angular momentum in kg m^2 s^-1
    :rtype: double
    """
    return L*G/c*mass_in_kg(M)**2

def angular_momentum_in_cgs(L,M):
    """
    Converts angular momentum in geometrized units to g cm^2 s^-1

    :param L: angular momentum in multiples of M^2
    :type L: double
    :param M: mass of the primary body in solar masses
    :type M: double

    :return: angular momentum in g cm^2 s^-1
    :rtype: double
    """
    return angular_momentum_in_mks(L,M)*1e7

def carter_constant_in_mks(Q,M):
    """
    Converts Carter constant in geometrized units to kg^2 m^4 s^-2

    :param Q: Carter constant in multiples of M^4
    :type Q: double
    :param M: mass of the primary body in solar masses
    :type M: double
    """
    return Q*G**2/c**2*mass_in_kg(M)**4

def carter_constant_in_cgs(Q,M):
    """
    Converts Carter constant in geometrized units to g^2 cm^4 s^-2

    :param Q: Carter constant in multiples of M^4
    :type Q: double
    :param M: mass of the primary body in solar masses
    :type M: double
    """
    return carter_constant_in_mks(Q,M)*1e14

def frequency_in_Hz(f,M):
    """
    Converts frequency in geometrized units to hertz

    :param f: frequency in multiples of M^-1
    :type f: _type_
    :param M: _description_
    :type M: _type_
    """
    return f*c**3/G/mass_in_kg(M)

def frequency_in_mHz(f,M):
    """
    Converts frequency in geometrized units to millihertz

    :param f: frequency in multiples of M^-1
    :type f: double
    :param M: mass of the primary body in solar masses
    :type M: double

    :return: frequency in millihertz
    :rtype: double
    """
    return frequency_in_Hz(f,M)*1e3



import unittest
import numpy as np
from kerrgeopy.bound_solutions import *
from kerrgeopy.frequencies_from_constants import _radial_roots, _polar_roots
from pathlib import Path

THIS_DIR = Path(__file__).parent

DATA_DIR = THIS_DIR.parent / "tests/data"

class TestBoundSolutions(unittest.TestCase):
    def test_random(self):
        components = ["t_r","t_theta","phi_r","phi_theta"]
        orbit_values = np.genfromtxt(DATA_DIR / "orbit_values.txt", delimiter=",")
        orbit_times = np.genfromtxt(DATA_DIR / "orbit_times.txt", delimiter=",")

        for i, orbit in enumerate(orbit_values):
            mathematica_trajectory = np.genfromtxt(DATA_DIR / f"geodesics/trajectory{i}.txt", delimiter=",")
            
            a,p,e,x = orbit
            constants = constants_of_motion(*orbit)
            radial_roots = _radial_roots(a,p,e,constants)
            polar_roots = _polar_roots(a,x,constants)
            r, t_r, phi_r = radial_solutions(a,constants,radial_roots)
            theta, t_theta, phi_theta = polar_solutions(a,constants,polar_roots)

            python_trajectory = np.transpose(
                np.apply_along_axis(lambda x: np.array([t_r(x),t_theta(x),phi_r(x),phi_theta(x)]),0,orbit_times)
                )
            
            for j, component in enumerate(components):
                with self.subTest(i=i,
                                  component=component,
                                  params="a = {}, p = {}, e = {}, x = {}".format(*orbit),
                                  diff=np.max(np.abs(mathematica_trajectory[:,j]-python_trajectory[:,j]))
                                  ):
                    self.assertTrue(np.allclose(mathematica_trajectory[:,j],python_trajectory[:,j]))


from numpy import sin, cos, sqrt, pi
from numpy.polynomial import Polynomial
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
from matplotlib.animation import FFMpegWriter

class Orbit:
    r"""
    Class representing an orbit in Kerr spacetime.

    :param a: spin parameter
    :type a: double
    :param init_position: initial position of the orbit :math:`(t_0,r_0,\theta_0,\phi_0)`
    :type init_position: tuple(double,double,double,double)
    :param init_velocity: initial four-velocity of the orbit :math:`(\frac{dt}{d\tau}(0), \frac{dr}{d\tau}(0),\frac{d\theta}{d\tau}(0),\frac{d\phi}{d\tau}(0))`
    :type init_velocity: tuple(double,double,double,double)

    :ivar a: spin parameter
    :ivar init_position: initial position of the orbit :math:`(t_0,r_0,\theta_0,\phi_0)`
    :ivar E: dimensionless energy
    :ivar L: dimensionless angular momentum
    :ivar Q: dimensionless carter constant
    """
    def __init__(self,a,init_position,init_velocity):
        self.a = a
        self.init_position = init_position
        self.init_velocity = init_velocity

        t0, r0, theta0, phi0 = init_position
        dt0, dr0, dtheta0, dphi0 = init_velocity
        sigma = r0**2+a**2*cos(theta0)**2
        delta = r0**2-2*r0+a**2

        # solve for E and L by writing the t and phi equations as a matrix equation
        A = np.array(
            [[(r0**2+a**2)**2/delta-a**2*(1-cos(theta0)**2), a-a*(r0**2+a**2)/delta],
            [a*(r0**2+a**2)/delta-a, 1/(1-cos(theta0)**2)-a**2/delta]]
        )
        b = np.array([sigma*dt0,sigma*dphi0])
        E, L = np.linalg.solve(A,b)

        # solve for Q by substituting E and L back into the theta equation
        Q = ((sigma*dtheta0)**2 + cos(theta0)**2*(self.a**2*(1-self.E**2)*(1-cos(theta0)**2)+self.L**2))/(1-cos(theta0)**2)

        self.E, self.L, self.Q = E, L, Q

        # standard form of the radial polynomial R(r)
        R = Polynomial([-a**2*Q, 2*L**2+2*Q+2*a**2*E**2-4*a*E*L, a**2*E**2-L**2-Q-a**2, 2, E**2-1])
        radial_roots = R.roots()
        # get the real roots and the complex roots
        real_roots = np.sort(np.real(radial_roots[np.isreal(radial_roots)]))
        complex_roots = radial_roots[np.iscomplex(radial_roots)]

        r_minus = 1-sqrt(1-a**2)

        if len(complex_roots) == 4: raise ValueError("Not a physical orbit")
        if len(complex_roots) == 2:
            self.plunging = True
        else:
            r4, r3, r2, r1 = real_roots
            if (r2 < r_minus) & (r0 < r1) & (r0 > r2): 
                self.plunging = True
            elif (r2 > r_minus) & (r0 < r1) & (r0 > r2):
                self.plunging = False
            elif (r2 > r_minus) & (r0 < r3) & (r0 > r4):
                self.plunging = True
        
    # def trajectory(init_phases=(0,0,0,0)):
    #     pass

    def plot(self,lambda0=0, lambda1=20, elevation=30 ,azimuth=-60, initial_phases=(0,0,0,0), grid=True, axes=True, thickness=1):
        """
        Creates a plot of the orbit

        :param lambda0: starting mino time
        :type lambda0: double, optional
        :param lambda1: ending mino time
        :type lambda1: double, optional
        :param elevation: camera elevation angle in degrees
        :type elevation: double, optional
        :param azimuth: camera azimuthal angle in degrees
        :type azimuth: double, optional
        :param initial_phases: tuple of initial phases, defaults to (0,0,0,0)
        :type initial_phases: tuple, optional
        :param grid: if true, grid lines are shown on plot
        :type grid: bool, optional
        :param axes: if true, axes are shown on plot
        :type axes: bool, optional
        :param thickness: line thickness of the orbit
        :type thickness: double, optional

        :return: matplotlib figure and axes
        :rtype: matplotlib.figure.Figure, matplotlib.axes._subplots.AxesSubplot
        """
        lambda_range = lambda1 - lambda0
        point_density = 500
        num_pts = lambda_range*point_density
        time = np.linspace(lambda0,lambda1,num_pts)

        t, r, theta, phi = self.trajectory(initial_phases)

        # compute trajectory in cartesian coordinates
        trajectory_x = r(time)*sin(theta(time))*cos(phi(time))
        trajectory_y = r(time)*sin(theta(time))*sin(phi(time))
        trajectory_z = r(time)*cos(theta(time))
        trajectory = np.column_stack((trajectory_x,trajectory_y,trajectory_z))

        # create sphere with radius equal to event horizon radius
        event_horizon = 1+sqrt(1-self.a**2)
        u = np.linspace(0, 2 * np.pi, 100)
        v = np.linspace(0, np.pi, 100)
        x_sphere = event_horizon * np.outer(np.cos(u), np.sin(v))
        y_sphere = event_horizon * np.outer(np.sin(u), np.sin(v))
        z_sphere = event_horizon * np.outer(np.ones(np.size(u)), np.cos(v))

        # convert viewing angles to radians
        elevation_rad = elevation*pi/180
        azimuth_rad = azimuth*pi/180

        # https://matplotlib.org/stable/api/toolkits/mplot3d/view_angles.html
        view_plane_normal = [cos(elevation_rad)*cos(azimuth_rad),cos(elevation_rad)*sin(azimuth_rad),sin(elevation_rad)]
        # matplotlib has no ray tracer so points behind the black hole must be filtered out manually
        # for each trajectory point compute the component normal to the viewing plane
        normal_component = np.apply_along_axis(lambda x: np.dot(view_plane_normal,x),1,trajectory)
        # compute the projection of each trajectory point onto the viewing plane
        projection = trajectory-np.transpose(normal_component*np.transpose(np.broadcast_to(view_plane_normal,(num_pts,3))))
        # find points in front of the viewing plane or outside the event horizon when projected onto the viewing plane
        condition = (np.linalg.norm(trajectory,axis=1) > event_horizon) & ((normal_component >= 0) | (np.linalg.norm(projection,axis=1) > event_horizon))
        x_visible = trajectory_x[condition]
        y_visible = trajectory_y[condition]
        z_visible = trajectory_z[condition]

        fig = plt.figure()
        ax = fig.add_subplot(projection='3d')
        
        # plot black hole
        ax.plot_surface(x_sphere, y_sphere, z_sphere, color='black',shade=False)
        # plot orbit
        ax.scatter(x_visible,y_visible,z_visible,color="red",s=thickness)

        # set viewing angle
        ax.view_init(elevation,azimuth)
        # set equal aspect ratio and orthogonal projection
        ax.set_box_aspect([np.ptp(x_visible),np.ptp(y_visible),np.ptp(z_visible)])
        # https://matplotlib.org/stable/gallery/mplot3d/projections.html
        ax.set_proj_type('ortho')

        # turn off grid and axes if specified
        if not grid: ax.grid(False)
        if not axes: ax.axis("off")

        return fig, ax
    
    def is_visible(self,point,elevation,azimuth):
        """
        Determines if a point is visible from a given viewing angle or obscured by the black hole. 
        Viewing angles are defined as in https://matplotlib.org/stable/api/toolkits/mplot3d/view_angles.html and black hole is centered at the origin.

        :param point: cartersian coordinates of point to test
        :type point: array_like
        :param elevation: camera elevation angle in degrees
        :type elevation: double
        :param azimuth: camera azimuthal angle in degrees
        :type azimuth: double
        """
        # compute event horizon radius
        event_horizon = 1+sqrt(1-self.a**2)

        # convert viewing angles to radians
        elevation_rad = elevation*pi/180
        azimuth_rad = azimuth*pi/180

        # https://matplotlib.org/stable/api/toolkits/mplot3d/view_angles.html
        view_plane_normal = [cos(elevation_rad)*cos(azimuth_rad),cos(elevation_rad)*sin(azimuth_rad),sin(elevation_rad)]

        # compute the component normal to the viewing plane
        normal_component = np.dot(view_plane_normal,point)
        # compute the projection of the point onto the viewing plane
        projection = point-normal_component*view_plane_normal

        # test if point is outside the event horizon and either in front of the viewing plane or outside the event horizon when projected onto the viewing plane
        return True if (np.linalg.norm(point) > event_horizon) & ((normal_component >= 0) | (np.linalg.norm(projection) > event_horizon)) else False
    
    def animate(self,filename,lambda0=0, lambda1=10, elevation=30 ,azimuth=-60, initial_phases=(0,0,0,0), grid=True, axes=True, thickness=2, tail="long"):
        """
        Saves an animation of the orbit as an mp4 file

        :param filename: filename to save the animation to
        :type filename: str
        :param lambda0: starting mino time, defaults to 0
        :type lambda0: double, optional
        :param lambda1: ending mino time, defaults to 10
        :type lambda1: double, optional
        :param elevation: camera elevation angle in degrees, defaults to 30
        :type elevation: double, optional
        :param azimuth: camera azimuthal angle in degrees, defaults to -60
        :type azimuth: double, optional
        :param initial_phases: tuple of initial phases, defaults to (0,0,0,0)
        :type initial_phases: tuple, optional
        :param grid: sets visibility of the grid, defaults to True
        :type grid: bool, optional
        :param axes: sets visibility of axes, defaults to True
        :type axes: bool, optional
        :param thickness: thickness of the tail of the orbit, defaults to 2
        :type thickness: double, optional
        :param tail: sets the length of the tail (options are "short", "long" and "none"), defaults to "long"
        :type tail: str, optional
        """
        num_pts = int(2e3)
        time = np.linspace(lambda0,lambda1,num_pts)

        fig = plt.figure(figsize=(7,7))
        ax = fig.add_subplot(projection='3d')
        eh = 1+sqrt(1-self.a**2)

        body = ax.scatter([],[],[],c="black")
        tail = ax.scatter([],[],[],c="red",s=thickness)

        t, r, theta, phi = self.trajectory(initial_phases)

        ax.view_init(35,-60)
        x = r(time)*sin(theta(time))*cos(phi(time))
        y = r(time)*sin(theta(time))*sin(phi(time))
        z = r(time)*cos(theta(time))
        trajectory = np.column_stack((x,y,z))

        azimuth_rad = azimuth*pi/180
        elevation_rad = elevation*pi/180

        # https://matplotlib.org/stable/api/toolkits/mplot3d/view_angles.html
        view_plane_normal = [cos(elevation_rad)*cos(azimuth_rad),cos(elevation_rad)*sin(azimuth_rad),sin(elevation_rad)]

        # matplotlib has no ray tracer so points behind the black hole must be filtered out manually
        # for each trajectory point compute the component normal to the viewing plane
        normal_component = np.apply_along_axis(lambda x: np.dot(view_plane_normal,x),1,trajectory)
        # compute the projection of each trajectory point onto the viewing plane
        projection = trajectory-np.transpose(normal_component*np.transpose(np.broadcast_to(view_plane_normal,(num_pts,3))))
        # find points in front of the viewing plane or outside the event horizon when projected onto the viewing plane
        condition = (np.dot(trajectory,view_plane_normal) >= 0) | (np.linalg.norm(projection,axis=1) > eh)

        # create sphere with radius equal to event horizon radius
        u = np.linspace(0, 2 * np.pi, 100)
        v = np.linspace(0, np.pi, 100)
        x_sphere = eh * np.outer(np.cos(u), np.sin(v))
        y_sphere = eh * np.outer(np.sin(u), np.sin(v))
        z_sphere = eh * np.outer(np.ones(np.size(u)), np.cos(v))

        # plot black hole
        ax.plot_surface(x_sphere, y_sphere, z_sphere, color='black',shade=False,zorder=0)

        # set axis limits
        ax.set_xlim([x.min(),x.max()])
        ax.set_ylim([y.min(),y.max()])
        ax.set_zlim([z.min(),z.max()])

        # set equal aspect ratio and orthogonal projection
        ax.set_box_aspect([np.ptp(x),np.ptp(y),np.ptp(z)])
        # https://matplotlib.org/stable/gallery/mplot3d/projections.html
        ax.set_proj_type('ortho')

         # turn off grid and axes if specified
        if not grid: ax.grid(False)
        if not axes: ax.axis("off")

        def animate(i,body,tail):
            # adjust length of tail
            start = 0
            if tail == "short": start = max(0,i-50)
            elif tail == "none": start = i

            condition_slice = condition[start:i]
            body._offsets3d = ([x[i]],[y[i]],[z[i]])
            tail._offsets3d = (x[start:i][condition_slice],y[start:i][condition_slice],z[start:i][condition_slice])
            
        # save to file
        ani = FuncAnimation(fig,animate,num_pts,fargs=(body,tail))
        FFwriter = FFMpegWriter(fps=60)
        ani.save(filename, writer = FFwriter)


import unittest
import numpy as np
from kerrgeopy.constants_of_motion import *
from pathlib import Path

THIS_DIR = Path(__file__).parent

DATA_DIR = THIS_DIR.parent / "tests/data"

class TestConstants(unittest.TestCase):
    def test_extreme_kerr(self):
        with self.assertRaises(ValueError): constants_of_motion(1,12,0.5,0.5)
        with self.assertRaises(ValueError): energy(1,0.5,12,0.5)
        with self.assertRaises(ValueError): angular_momentum(1,0.5,12,0.5)
        with self.assertRaises(ValueError): carter_constant(1,0.5,12,0.5)

    def test_unstable(self):
        with self.assertRaises(ValueError): constants_of_motion(0.5,5,0.5,0.5)
        with self.assertRaises(ValueError): energy(0.5,5,0.5,0.5)
        with self.assertRaises(ValueError): angular_momentum(0.5,5,0.5,0.5)
        with self.assertRaises(ValueError): carter_constant(0.5,5,0.5,0.5)

    def test_invalid_arguments(self):
        with self.assertRaises(ValueError): constants_of_motion(2,5,0.5,0.5)
        with self.assertRaises(ValueError): energy(0.5,5,2,0.5)
        with self.assertRaises(ValueError): angular_momentum(0.5,5,0.5,2)
        with self.assertRaises(ValueError): carter_constant(-0.5,5,-0.5,0.5)

    def test_constants_random(self):
        values = np.genfromtxt(DATA_DIR / "const_values.txt",delimiter=",")
        mathematica_const_output = np.genfromtxt(DATA_DIR / "mathematica_const_output.txt")
        python_const_output = np.apply_along_axis(lambda x: constants_of_motion(*x),1,values)

        for i, params in enumerate(values):
            with self.subTest(i=i,params=params):
                self.assertTrue(np.allclose(mathematica_const_output[i],python_const_output[i]))


class TestSeparatrix(unittest.TestCase):
    def test_extreme_kerr(self):
        with self.assertRaises(ValueError): separatrix(1,0.5,0.5)

    def test_invalid_arguments(self):
        with self.assertRaises(ValueError): separatrix(2,-0.5,-0.5)

    def test_separatrix_random(self):
        sep_values = np.genfromtxt(DATA_DIR / "sep_values.txt",delimiter=",")
        mathematica_separatrix_output = np.genfromtxt(DATA_DIR / "mathematica_sep_output.txt")
        python_separatrix_output = np.apply_along_axis(lambda x: separatrix(*x),1,sep_values)
        for i, params in enumerate(sep_values):
            with self.subTest(i=i,params=params):
                self.assertTrue(np.allclose(python_separatrix_output[i],mathematica_separatrix_output[i]))


from scipy.special import ellipk, ellipe, elliprj, elliprf
from .constants_of_motion import *

def _ellippi(n,k):
    r"""
    Complete elliptic integral of the third kind defined as :math:`\Pi(n,k) = \int_0^{\frac{\pi}{2}} \frac{d\theta}{(1-n\sin^2{\theta})\sqrt{1-k^2\sin^2{\theta}}}`
    
    :type n: double
    :type k: double

    :rtype: double
    """
    # Note: sign of n is reversed from the definition in Fujita and Hikida

    # formula from https://en.wikipedia.org/wiki/Carlson_symmetric_form
    return elliprf(0,1-k**2,1)+1/3*n*elliprj(0,1-k**2,1,1-n)

def _radial_roots(a,p,e,constants):
    """
    Computes r1, r2, r3 and r4 as defined in equation 10 of Fujita and Hikida (arXiv:0906.1420)

    :param a: dimensionless spin of the black hole
    :type a: double
    :param p: orbital semi-latus rectum
    :type p: double
    :param e: orbital eccentricity
    :type e: double
    :param x: cosine of the orbital inclination
    :type x: tuple(double, double, double, double)
    :param constants: dimensionless constants of motion for the orbit
    :type constants: tuple(double, double, double)

    :return: tuple containing the four roots of the radial equation
    :rtype: tuple(double, double, double, double)
    """
    E, L, Q = constants
    
    r1 = p/(1-e)
    r2 = p/(1+e)
    
    A_plus_B = 2/(1-E**2)-r1-r2
    AB = a**2*Q/(r1*r2*(1-E**2))
    
    r3 = (A_plus_B+sqrt(A_plus_B**2-4*AB))/2
    r4 = AB/r3
    
    return r1, r2, r3, r4

def _polar_roots(a,x,constants):
    r"""
    Computes epsilon_0, z_minus and z_plus as defined in equation 10 of Fujita and Hikida (arXiv:0906.1420)

    :param a: dimensionless spin of the black hole
    :type a: double
    :param x: cosine of the orbital inclination
    :type x: tuple(double, double, double)
    :param constants: dimensionless constants of motion for the orbit
    :type constants: tuple(double, double, double)

    :return: tuple of the form :math:`(\epsilon_0, z_-, z_+)`
    :rtype: tuple(double, double, double, double)
    """
    E, L, Q = constants
    epsilon0 = a**2*(1-E**2)/L**2
    z_minus = 1-x**2
    #z_plus = a**2*(1-E**2)/(L**2*epsilon0)+1/(epsilon0*(1-z_minus))
    # simplified using definition of carter constant
    z_plus = nan if a == 0 else 1+1/(epsilon0*(1-z_minus)) 
    
    return z_minus, z_plus

def r_frequency_from_constants(constants,radial_roots):
    """
    Computes the frequency of motion in r in Mino time using the method derived in Fujita and Hikida (arXiv:0906.1420)

    :param constants: dimensionless constants of motion for the orbit in the form :math:`(E,L,Q)`
    :type constants: tuple(double, double, double)
    :param radial_roots: tuple containing the four roots of the radial equation
    :type radial_roots: tuple(double, double, double, double)

    :rtype: double
    """
    E, L, Q = constants

    r1,r2,r3,r4 = radial_roots
    # equation 13
    k_r = sqrt((r1-r2)*(r3-r4)/((r1-r3)*(r2-r4)))
    # equation 15
    return pi*sqrt((1-E**2)*(r1-r3)*(r2-r4))/(2*ellipk(k_r**2))

def theta_frequency_from_constants(a,constants,radial_roots,polar_roots):
    """
    Computes the frequency of motion in theta in Mino time using the method derived in Fujita and Hikida (arXiv:0906.1420)
    
    :param a: dimensionless spin of the black hole
    :type a: double
    :param constants: dimensionless constants of motion for the orbit in the form :math:`(E,L,Q)`
    :type constants: tuple(double, double, double)
    :param radial_roots: tuple containing the four roots of the radial polynomial
    :type radial_roots: tuple(double, double, double, double)
    :param polar_roots: tuple containing the roots of the polar equation :math:`(\epsilon_0, z_-, z_+)`
    :type polar_roots: tuple(double, double)
    
    :rtype: double
    """
    r1, r2, r3, r4 = radial_roots
    z_minus, z_plus = polar_roots
    E, L, Q = constants

    # Schwarzschild case
    if a == 0:
        p = 2*r1*r2/(r1+r2)
        e = (r1-r2)/(r1+r2)
        return p/sqrt(-3-e**2+p)*(sign(L) if e == 0 else 1)
    
    
    # equation 13
    k_theta = sqrt(z_minus/z_plus)
    # simplified form of epsilon0*z_plus
    e0zp = (a**2*(1-E**2)*(1-z_minus)+L**2)/(L**2*(1-z_minus))
    
    # equation 15
    return pi*L*sqrt(e0zp)/(2*ellipk(k_theta**2))

def phi_frequency_from_constants(a,constants,radial_roots,polar_roots,upsilon_r=None,upsilon_theta=None):
    """
    Computes the frequency of motion in phi in Mino time using the method derived in Fujita and Hikida (arXiv:0906.1420)
    
    :param a: dimensionless spin of the black hole
    :type a: double
    :param constants: dimensionless constants of motion for the orbit in the form :math:`(E,L,Q)`
    :type constants: tuple(double, double, double)
    :param radial_roots: tuple containing the four roots of the radial polynomial
    :type radial_roots: tuple(double, double, double, double)
    :param polar_roots: tuple containing the roots of the polar equation :math:`(\epsilon_0, z_-, z_+)`
    :type polar_roots: tuple(double, double)
    :param upsilon_r: Mino frequency of motion in r can be passed in to speed computation if it is already known
    :type upsilon_r: double, optional
    :param upsilon_theta: Mino frequency of motion in theta can be passed in to speed computation if it is already known
    :type upsilon_theta: double, optional
    
    :rtype: double
    """

    E, L, Q = constants
    r1,r2,r3,r4 = radial_roots
    z_minus, z_plus = polar_roots

    # Schwarzschild case
    if a == 0:
        p = 2*r1*r2/(r1+r2)
        e = (r1-r2)/(r1+r2)
        return sign(L)*p/sqrt(-3-e**2+p)
    
    # compute frequencies if they are not passed in
    if upsilon_r is None: upsilon_r = r_frequency_from_constants(a,constants,radial_roots)
    if upsilon_theta is None: upsilon_theta = theta_frequency_from_constants(a,constants,radial_roots,polar_roots)
    
    # simplified form of epsilon0*z_plus
    e0zp = (a**2*(1-E**2)*(1-z_minus)+L**2)/(L**2*(1-z_minus))

    r_plus = 1+sqrt(1-a**2)
    r_minus = 1-sqrt(1-a**2)
    
    h_plus = (r1-r2)*(r3-r_plus)/((r1-r3)*(r2-r_plus))
    h_minus = (r1-r2)*(r3-r_minus)/((r1-r3)*(r2-r_minus))
    
    # equation 13
    k_theta = sqrt(z_minus/z_plus)
    k_r = sqrt((r1-r2)*(r3-r4)/((r1-r3)*(r2-r4)))
    
    # equation 21
    return  2*upsilon_theta/(pi*sqrt(e0zp))*_ellippi(z_minus,k_theta) \
            + 2*a*upsilon_r/(pi*(r_plus-r_minus)*sqrt((1-E**2)*(r1-r3)*(r2-r4))) \
            * ((2*E*r_plus-a*L)/(r3-r_plus)*(ellipk(k_r**2)-(r2-r3)/(r2-r_plus)*_ellippi(h_plus,k_r)) \
               - (2*E*r_minus-a*L)/(r3-r_minus)*(ellipk(k_r**2)-(r2-r3)/(r2-r_minus)*_ellippi(h_minus,k_r))
              )

def gamma_from_constants(a,constants,radial_roots,polar_roots,upsilon_r=None,upsilon_theta=None):
    """
    Computes the average rate at which observer time accumulates in Mino time using the method derived in Fujita and Hikida (arXiv:0906.1420)
    
    :param a: dimensionless spin of the black hole
    :type a: double
    :param constants: dimensionless constants of motion for the orbit in the form :math:`(E,L,Q)`
    :type constants: tuple(double, double, double)
    :param radial_roots: tuple containing the four roots of the radial polynomial
    :type radial_roots: tuple(double, double, double, double)
    :param polar_roots: tuple containing the roots of the polar equation :math:`(\epsilon_0, z_-, z_+)`
    :type polar_roots: tuple(double, double)
    :param upsilon_r: Mino frequency of motion in r can be passed in to speed computation if it is already known
    :type upsilon_r: double, optional
    :param upsilon_theta: Mino frequency of motion in theta can be passed in to speed computation if it is already known
    :type upsilon_theta: double, optional
    
    :rtype: double
    """
    r1,r2,r3,r4 = radial_roots
    z_minus, z_plus = polar_roots

    e = (r1-r2)/(r1+r2)
    # marginally bound case
    if e == 1:
        return inf
    
    E, L, Q = constants
    
    epsilon0 =  a**2*(1-E**2)/L**2
    # simplified form of a**2*sqrt(z_plus/epsilon0)
    a2sqrt_zp_over_e0 = L**2/((1-E**2)*sqrt(1-z_minus)) if a == 0 else a**2*z_plus/sqrt(epsilon0*z_plus)
    
    # compute frequencies if they are not passed in
    if upsilon_r is None: upsilon_r = r_frequency_from_constants(a,constants,radial_roots)
    if upsilon_theta is None: upsilon_theta = theta_frequency_from_constants(a,constants,radial_roots,polar_roots)
    
    r_plus = 1+sqrt(1-a**2)
    r_minus = 1-sqrt(1-a**2)
    
    h_r = (r1-r2)/(r1-r3)
    h_plus = (r1-r2)*(r3-r_plus)/((r1-r3)*(r2-r_plus))
    h_minus = (r1-r2)*(r3-r_minus)/((r1-r3)*(r2-r_minus))
    
    # equation 13
    k_theta = 0 if a == 0 else sqrt(z_minus/z_plus)
    k_r = sqrt((r1-r2)*(r3-r4)/((r1-r3)*(r2-r4)))
    
    # equation 21
    return 4*E + 2*a2sqrt_zp_over_e0*E*upsilon_theta*(ellipk(k_theta**2)-ellipe(k_theta**2))/(pi*L) \
            + 2*upsilon_r/(pi*sqrt((1-E**2)*(r1-r3)*(r2-r4))) \
            * (E/2*((r3*(r1+r2+r3)-r1*r2)*ellipk(k_r**2) + (r2-r3)*(r1+r2+r3+r4)*_ellippi(h_r,k_r)+ (r1-r3)*(r2-r4)*ellipe(k_r**2)) \
               + 2*E*(r3*ellipk(k_r**2)+(r2-r3)*_ellippi(h_r,k_r))\
               +2/(r_plus-r_minus)*(((4*E-a*L)*r_plus-2*a**2*E)/(r3-r_plus)*(ellipk(k_r**2)-(r2-r3)/(r2-r_plus)*_ellippi(h_plus,k_r)) \
                                    -((4*E-a*L)*r_minus-2*a**2*E)/(r3-r_minus)*(ellipk(k_r**2)-(r2-r3)/(r2-r_minus)*_ellippi(h_minus,k_r))
                                   )
              )

def mino_frequencies_from_constants(a,constants,radial_roots,polar_roots):
    r"""
    Computes frequencies of orbital motion in Mino time using the method derived in Fujita and Hikida (arXiv:0906.1420)

    :param a: dimensionless spin of the black hole (must satisfy -1 < a < 1)
    :type a: double
    :param constants: dimensionless constants of motion for the orbit in the form :math:`(E,L,Q)`
    :type constants: tuple(double, double, double)
    :param radial_roots: tuple containing the four roots of the radial polynomial
    :type radial_roots: tuple(double, double, double, double)
    :param polar_roots: tuple containing the roots of the polar equation :math:`(\epsilon_0, z_-, z_+)`
    :type polar_roots: tuple(double, double)

    :return: tuple of the form :math:`(\Upsilon_r, \Upsilon_\theta, \Upsilon_\phi, \Gamma)`
    :rtype: tuple(double, double, double, double)
    """
    upsilon_r = r_frequency_from_constants(a,constants,radial_roots)
    upsilon_theta = theta_frequency_from_constants(a,constants,radial_roots,polar_roots)
    upsilon_phi = phi_frequency_from_constants(a,constants,radial_roots,polar_roots,upsilon_r,upsilon_theta)
    Gamma = gamma_from_constants(a,constants,radial_roots,polar_roots,upsilon_r,upsilon_theta)

    return upsilon_r, abs(upsilon_theta), upsilon_phi, Gamma
    
def observer_frequencies_from_constants(a,constants,radial_roots,polar_roots):
    r"""
    Computes frequencies of orbital motion in Boyer-Lindquist time using the method derived in Fujita and Hikida (arXiv:0906.1420)

    :param a: dimensionless spin of the black hole (must satisfy -1 < a < 1)
    :type a: double
    :param p: orbital semi-latus rectum
    :type p: double
    :param e: orbital eccentricity (must satisfy 0 <= e < 1)
    :type e: double
    :param x: cosine of the orbital inclination (must satisfy 0 < x^2 <= 1)
    :type x: double

    :return: tuple of the form :math:`(\Omega_r, \Omega_\theta, \Omega_\phi)`
    :rtype: tuple(double, double, double)
    """
    upsilon_r = r_frequency_from_constants(a,constants,radial_roots)
    upsilon_theta = theta_frequency_from_constants(a,constants,radial_roots,polar_roots)
    upsilon_phi = phi_frequency_from_constants(a,constants,radial_roots,polar_roots,upsilon_r,upsilon_theta)
    Gamma = gamma_from_constants(a,constants,radial_roots,polar_roots,upsilon_r,upsilon_theta)

    return upsilon_r/Gamma, abs(upsilon_theta)/Gamma, upsilon_phi/Gamma


__all__ = ["constants_of_motion","frequencies","bound_orbit","plunging_orbit"]
from kerrgeopy import *
from kerrgeopy.frequencies import *
from kerrgeopy.constants_of_motion import *
from kerrgeopy.bound_orbit import BoundOrbit
from kerrgeopy.plunging_orbit import PlungingOrbit


from numpy import sort_complex, sqrt, arctan, arctan2, arccos, log, sin, cos, pi
from numpy.polynomial import Polynomial
import numpy as np
from scipy.special import ellipj, ellipeinc, ellipk
from .bound_solutions import _ellippiinc

def _plunging_radial_roots(a,E,L,Q):
    """
    Computes the radial roots for a plunging orbit. The roots are sorted in ascending order with real roots first.
    If all roots are real, the roots are sorted such that the motion is between the first two roots.

    :param a: dimensionless spin parameter
    :type a: double
    :param E: dimensionless energy
    :type E: double
    :param L: dimensionless angular momentum
    :type L: double
    :param Q: dimensionless Carter constant
    :type Q: double

    :return: tuple of radial roots
    :rtype: tuple(double,double,double,double)
    """
    # standard form of the radial polynomial R(r)
    R = Polynomial([-a**2*Q, 2*L**2+2*Q+2*a**2*E**2-4*a*E*L, a**2*E**2-L**2-Q-a**2, 2, E**2-1])
    roots = R.roots()
    # get the real roots and the complex roots
    real_roots = np.sort(np.real(roots[np.isreal(roots)]))
    complex_roots = roots[np.iscomplex(roots)]

    r_minus = 1-sqrt(1-a**2)

    if len(real_roots) == 4:
        # if there are three roots outside the event horizon swap r1/r3 and r2/r4
        if real_roots[1] > r_minus:
            r1 = real_roots[1]
            r2 = real_roots[0]
            r3 = real_roots[3]
            r4 = real_roots[2]
        else:
            r1, r2, r3, r4 = real_roots

    elif len(real_roots) == 2:
        r1, r2 = real_roots
        r3, r4 = complex_roots

    return r1, r2, r3, r4

def plunging_mino_frequencies(a,E,L,Q):
    r"""
    Computes the radial and polar mino frequencies for a plunging orbit.

    :param a: dimensionless spin parameter
    :type a: double
    :param E: dimensionless energy
    :type E: double
    :param L: dimensionless angular momentum
    :type L: double
    :param Q: dimensionless Carter constant
    :type Q: double

    :return: radial and polar mino frequencies :math:`(\Upsilon_r,\Upsilon_\theta)`
    :rtype: tuple(double,double)
    """
    r1, r2, r3, r4 = _plunging_radial_roots(a,E,L,Q)
    rho_r = np.real(r3)
    rho_i = np.imag(r4)

    # equation 42
    A = sqrt((r2-rho_r)**2+rho_i**2)
    B = sqrt((r1-rho_r)**2+rho_i**2)
    k_r = sqrt(((r2-r1)**2-(A-B)**2)/(4*A*B))

    # equation 52
    upsilon_r = pi*sqrt(A*B*(1-E**2))/(2*ellipk(k_r**2))

    # equation 14
    z1 = sqrt(1/2*(1+(L**2+Q)/(a**2*(1-E**2))-sqrt((1+(L**2+Q)/(a**2*(1-E**2)))**2-4*Q/(a**2*(1-E**2)))))
    z2 = sqrt(a**2*(1-E**2)/2*(1+(L**2+Q)/(a**2*(1-E**2))+sqrt((1+(L**2+Q)/(a**2*(1-E**2)))**2-4*Q/(a**2*(1-E**2)))))
    k_theta = a*sqrt(1-E**2)*z1/z2

    # equation 53
    upsilon_theta = pi*ellipk(k_theta**2)/(2*z2)

    return upsilon_r, upsilon_theta

def plunging_radial_integrals(a,E,L,Q):
    r"""
    Computes the radial integrals :math:`I_r`, :math:`I_{r^2}` and :math:`I_{r_\pm}` defined in equation 39 of Dyson and van de Meent (arXiv:2302.03704) as a function of the radial phase.

    :param a: dimensionless spin parameter
    :type a: double
    :param E: dimensionless energy
    :type E: double
    :param L: dimensionless angular momentum
    :type L: double
    :param Q: dimensionless Carter constant
    :type Q: double

    :return: radial integrals :math:`(I_r,I_{r^2},I_{r_+},I_{r_-})`
    :rtype: tuple(function,function,function,function)
    """
    r1, r2, r3, r4 = _plunging_radial_roots(a,E,L,Q)
    rho_r = np.real(r3)
    rho_i = np.imag(r4)

    # inner and outer horizons
    r_plus = 1+sqrt(1-a**2)
    r_minus = 1-sqrt(1-a**2)

    # equation 42
    A = sqrt((r2-rho_r)**2+rho_i**2)
    B = sqrt((r1-rho_r)**2+rho_i**2)
    f = 4*A*B/(A-B)**2
    k_r = sqrt(((r2-r1)**2-(A-B)**2)/(4*A*B))

    upsilon_r, upsilon_theta = plunging_mino_frequencies(a,E,L,Q)

    # equation 48
    D_plus = sqrt(4*A*B*(r2-r_plus)*(r_plus-r1))/(A*(r_plus-r1)+B*(r2-r_plus))
    D_minus = sqrt(4*A*B*(r2-r_minus)*(r_minus-r1))/(A*(r_minus-r1)+B*(r2-r_minus))

    def I_r(q_r):
        # equation 43
        sn, cn, dn, xi_r = ellipj(2*ellipk(k_r**2)*q_r/pi,k_r**2)
        mino_time = q_r/upsilon_r
        # equation 46
        return (A*r1-B*r2)/(A-B)*mino_time - 1/sqrt(1-E**2)*arctan((r2-r1)*sn/(2*sqrt(A*B*dn))) \
                + (A+B)*(r2-r1)/(2*(A-B)*sqrt(A*B*(1-E**2)))*_ellippiinc(xi_r,-1/f,k_r)
    
    def I_r2(q_r):
        # equation 43
        sn, cn, dn, xi_r = ellipj(2*ellipk(k_r**2)*q_r/pi,k_r**2)
        mino_time = q_r/upsilon_r
        # equation 47
        return (A*r1**2-B*r2**2)/(A-B)*mino_time + sqrt(A*B)/sqrt(1-E**2)*ellipeinc(xi_r,k_r**2) \
                - (A+B)*(A**2+2*r1**2-B**2-2*r2**2)/(4*(A-B)*sqrt((1-E**2)*A*B))*_ellippiinc(xi_r,-1/f,k_r) \
                - sqrt(A*B)*(A+B-(A-B)*cn)*sn*dn/((A-B)*sqrt(1-E**2)*(f+sn**2)) \
                + (A**2+2*r1**2-B**2-2*r2**2)/(4*(r2-r1)*sqrt(1-E**2))*arctan2(2*sn*dn*sqrt(f*(1+f*k_r**2)),f-(1+2*f*k_r**2)*sn**2)
        
    def I_r_plus(q_r):
        # equation 43
        sn, cn, dn, xi_r = ellipj(2*ellipk(k_r**2)*q_r/pi,k_r**2)
        mino_time = q_r/upsilon_r
        # equation 48
        return (A-B)*mino_time/(A*(r1-r_plus)-B*(r2-r_plus)) \
            + (r2-r1)*(A*(r1-r_plus)+B*(r2-r_plus))/(2*sqrt(A*B*(1-E**2))*(r_plus-r1)*(r2-r_plus)*(A*(r1-r_plus)-B*(r2-r_plus)))*_ellippiinc(xi_r,1/D_plus**2,k_r) \
            - (sqrt(r2-r1)*log(
                ((D_plus*sqrt(1-D_plus**2*k_r**2)+dn*sn)**2 + (k_r*(D_plus**2-sn**2))**2) /
                ((D_plus*sqrt(1-D_plus**2*k_r**2)-dn*sn)**2 + (k_r*(D_plus**2-sn**2))**2)
                )  /
                (4*sqrt((1-E**2)*(r2-r_plus)*(r_plus-r1))*sqrt(A**2*(r_plus-r1)-(r2-r_plus)*(r1**2-B**2+r2*r_plus-r1*(r2+r_plus))))
            )
    
    def I_r_minus(q_r):
        # equation 43
        sn, cn, dn, xi_r = ellipj(2*ellipk(k_r**2)*q_r/pi,k_r**2)
        mino_time = q_r/upsilon_r
        # equation 48
        return (A-B)*mino_time/(A*(r1-r_minus)-B*(r2-r_minus)) \
            + (r2-r1)*(A*(r1-r_minus)+B*(r2-r_minus))/(2*sqrt(A*B*(1-E**2))*(r_minus-r1)*(r2-r_minus)*(A*(r1-r_minus)-B*(r2-r_minus)))*_ellippiinc(xi_r,1/D_minus**2,k_r) \
            - (sqrt(r2-r1)*log(
                ((D_minus*sqrt(1-D_minus**2*k_r**2)+dn*sn)**2 + (k_r*(D_minus**2-sn**2))**2) /
                ((D_minus*sqrt(1-D_minus**2*k_r**2)-dn*sn)**2 + (k_r*(D_minus**2-sn**2))**2)
                )  /
                (4*sqrt((1-E**2)*(r2-r_minus)*(r_minus-r1))*sqrt(A**2*(r_minus-r1)-(r2-r_minus)*(r1**2-B**2+r2*r_minus-r1*(r2+r_minus))))
            )

    return I_r, I_r2, I_r_plus, I_r_minus

def plunging_radial_solutions_complex(a,E,L,Q):
    r"""
    Computes the radial solutions :math:`r(q_r), t_r(q_r), \phi_r(q_r)` from equation 50 and 51 of Dyson and van de Meent (arXiv:2302.03704) for the case of two complex radial roots.

    :param a: dimensionless spin parameter
    :type a: double
    :param E: dimensionless energy
    :type E: double
    :param L: dimensionless angular momentum
    :type L: double
    :param Q: dimensionless Carter constant
    :type Q: double

    :return: tuple of radial solutions :math:`(r(q_r), t_r(q_r), \phi_r(q_r))`
    :rtype: tuple(function, function, function)
    """
    roots = _plunging_radial_roots(a,E,L,Q)
    if np.iscomplex(roots).sum() != 2: raise ValueError("There should be two complex roots")

    r1, r2, r3, r4 = roots
    rho_r = np.real(r3)
    rho_i = np.imag(r4)

    # inner and outer horizons
    r_plus = 1+sqrt(1-a**2)
    r_minus = 1-sqrt(1-a**2)

    # equation 42
    A = sqrt((r2-rho_r)**2+rho_i**2)
    B = sqrt((r1-rho_r)**2+rho_i**2)
    f = 4*A*B/(A-B)**2
    k_r = sqrt(((r2-r1)**2-(A-B)**2)/(4*A*B))

    upsilon_r, upsilon_theta = plunging_mino_frequencies(a,E,L,Q)

    I_r, I_r2, I_r_plus, I_r_minus = plunging_radial_integrals(a,E,L,Q)

    def r(q_r):
        # equation 43
        sn, cn, dn, xi_r = ellipj(2*ellipk(k_r**2)*q_r/pi,k_r**2)
        # equation 49
        return ((A-B)*(A*r1-B*r2)*sn**2+2*A*B*(r1+r2)-2*A*B*(r2-r1)*cn)/(4*A*B+(A-B)**2*sn**2)
    
    def t_r(q_r):
        mino_time = q_r/upsilon_r
        # equation 41
        return E*(r_plus**2+r_minus**2+r_plus*r_minus+2*a**2)*mino_time + \
            E*(I_r2(q_r)+I_r(q_r)*(r_minus+r_plus)) + \
            ((r_minus**2+a**2)*(E*(r_minus**2+a**2)-a*L)/(r_minus-r_plus)*I_r_minus(q_r) + 
             (r_plus**2+a**2)*(E*(r_plus**2+a**2)-a*L)/(r_plus-r_minus)*I_r_plus(q_r)
            ) - a*L*mino_time

    def phi_r(q_r):
        mino_time = q_r/upsilon_r
        # equation 40
        return a*(
            (E*(r_minus**2+a**2)-a*L)/(r_minus-r_plus)*I_r_minus(q_r) +
            (E*(r_plus**2+a**2)-a*L)/(r_plus-r_minus)*I_r_plus(q_r)
            ) + a*E*mino_time
    
    return r, t_r, phi_r

def plunging_polar_solutions(a,E,L,Q):
    r"""
    Computes the polar solutions :math:`\theta(q_\theta), t_\theta(q_\theta), \phi_\theta(q_\theta)` from equation 33 and 37 of Dyson and van de Meent (arXiv:2302.03704).

    :param a: dimensionless spin parameter
    :type a: double
    :param E: dimensionless energy
    :type E: double
    :param L: dimensionless angular momentum
    :type L: double
    :param Q: dimensionless Carter constant
    :type Q: double
    
    :return: tuple of polar solutions :math:`(\theta(q_\theta), t_\theta(q_\theta), \phi_\theta(q_\theta))`
    :rtype: tuple(function, function, function)
    """
    z1 = sqrt(1/2*(1+(L**2+Q)/(a**2*(1-E**2))-sqrt((1+(L**2+Q)/(a**2*(1-E**2)))**2-4*Q/(a**2*(1-E**2)))))
    z2 = sqrt(a**2*(1-E**2)/2*(1+(L**2+Q)/(a**2*(1-E**2))+sqrt((1+(L**2+Q)/(a**2*(1-E**2)))**2-4*Q/(a**2*(1-E**2)))))
    k_theta = a*sqrt(1-E**2)*z1/z2

    upsilon_r, upsilon_theta = plunging_mino_frequencies(a,E,L,Q)

    def theta(q_theta):
        mino_time = q_theta/upsilon_theta
        # equation 28
        sn, cn, dn, xi_theta = ellipj(z2*mino_time,k_theta**2)
        # equation 27
        return arccos(z1*sn)
    
    def t_theta(q_theta):
        mino_time = q_theta/upsilon_theta
        # equation 28
        sn, cn, dn, xi_theta = ellipj(z2*mino_time,k_theta**2)
        # equation 35
        return E/(1-E**2)*((z2**2-a**2*(1-E**2))*mino_time-z2*ellipeinc(xi_theta,k_theta**2))
    
    def phi_theta(q_theta):
        mino_time = q_theta/upsilon_theta
        # equation 28
        sn, cn, dn, xi_theta = ellipj(z2*mino_time,k_theta**2)
        # equation 31
        return L/z2*_ellippiinc(xi_theta,z1**2,k_theta**2)
    
    return theta, t_theta, phi_theta
